/*
** Copyright (C) 2002-2004 Daniel Caujolle-Bert <segfault@club-internet.fr>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include "readline.h"
#include "tilde.h"

extern char **environ;
extern int errno;

#include "common.h"
#include "playlist.h"

extern prompt_state_t prompt_state[];

char *toxine_get_yesno_string(uint32_t val) {
  static char *yesno[] =  { "No", "Yes" };

  return ((val > 0) ? yesno[1] : yesno[0]);
}

char *toxine_get_fourcc_string(uint32_t f) {
  static char fcc[5];
  
  memset(&fcc, 0, sizeof(fcc));
  
  /* Should we take care about endianess ? */
  fcc[0] = f     | 0xFFFFFF00;
  fcc[1] = f>>8  | 0xFFFFFF00;
  fcc[2] = f>>16 | 0xFFFFFF00;
  fcc[3] = f>>24 | 0xFFFFFF00;
  fcc[4] = 0;
  
  if(f <= 0xFFFF)
    sprintf(fcc, "0x%x", f);
  
  if((fcc[0] == 'm') && (fcc[1] == 's')) {
    if((fcc[2] = 0x0) && (fcc[3] == 0x55)) {
      *(uint32_t *) fcc = 0x33706d2e; /* Force to '.mp3' */
    }
  }
  
  return &fcc[0];
}

/*
 * cleanup the str string, take care about '
 */
char *toxine_atoa(char *str) {
  char *pbuf;
  int   quote = 0, dblquote = 0;

  pbuf = str;
  
  while(*pbuf == ' ') 
    pbuf++;

  if(*pbuf == '\'')
    quote = 1;
  else if(*pbuf == '"')
    dblquote = 1;
  
  pbuf = str;

  while(*pbuf != '\0')
    pbuf++;
  
  if(pbuf > str)
    pbuf--;

  while((pbuf > str) && (*pbuf == '\r' || *pbuf == '\n')) {
    *pbuf = '\0';
    pbuf--;
  }

  while((pbuf > str) && (*pbuf == ' ')) {
    *pbuf = '\0';
    pbuf--;
  }
  
  if((quote && (*pbuf == '\'')) || (dblquote && (*pbuf == '"'))) {
    *pbuf = '\0';
    pbuf--;
  }
  
  pbuf = str;

  while(*pbuf == ' ' || *pbuf == '\t')
    pbuf++;
  
  if((quote && (*pbuf == '\'')) || (dblquote && (*pbuf == '"')))
    pbuf++;
  
  return pbuf;
}

int toxine_confirm(char *message) {
  char  *line;
  char  prompt[strlen(HELP_PREFIX) + strlen(message) + 1];
  
  sprintf(prompt, "%s%s", HELP_PREFIX, message);
  if((line = readline(prompt)) == NULL) {
    perr("%s(%d): readline() failed: %s\n", __tox_func__, __LINE__,
	 strerror(errno));
    exit(1);
  }
  
  if (*line) {
    line = toxine_atoa(line);
    
    if(!strncasecmp(line, "yes", strlen(line))) {
      toxine_free(line);
      return 1;
    }
    
    toxine_free(line);
  }
  return 0;
}

void toxine_draw_bar(char *title, int min, int max, int val) {
  float    _val = (int) val;
  float    _min = (int) min;
  float    _max = (int) max;
  float    range, percent;
  int      i, screen_rows, screen_cols;
  
  if(max <= min)
    _max = (int) (min + 1);
  if(min >= max)
    _min = (int) (max - 1);
  if(val > max)
    _val = (int) max;
  if(val < min)
    _val = (int) min;

  if(title)
    pinfo("%s: (%%%.2f) %d > [%d] < %d\n", title,
	  (((_val + -_min) / _max) * 100.0), min, val, max);
  else
    pinfo("- (%%%.2f) %d > [%d] < %d\n", (((_val + -_min) / _max) * 100.0), min, val, max);
  
  rl_get_screen_size(&screen_rows, &screen_cols);
  
  range = (_max + -_min) / (screen_cols - 20);
  percent = (_val + -_min) / range;
  
  pinfo("|");
  for(i = 0; i <= (screen_cols - 20); i++) {

    if(i == (int) percent)
      pout("*");
    else
      pout("-");

  }
  pout("|\n");
  pinfo(".\n");
}

/*
 * Update the prompt, reflect the current xine_engine state.
 */
void toxine_update_prompt(toxine_t *tox) {
  int i = 0, s = -1; 

  if(tox->xine_state & XINE_STREAM)
    s = xine_get_status(tox->stream);
  
  while(prompt_state[i].state != NULL) {
    if(prompt_state[i].xine_state == s) {
      sprintf(tox->prompt, PROMPT, prompt_state[i].state);
      return;
    }
    i++;
  }

  sprintf(tox->prompt, PROMPT, prompt_state[i-1].state);
}

/*
 * return a boolean value from a given string.
 */
int toxine_get_bool_value(const char *val) {
  static struct {
    const char *str;
    int value;
  } bools[] = {
    { "1",     1 }, { "true",  1 }, { "yes",   1 }, { "on",    1 },
    { "0",     0 }, { "false", 0 }, { "no",    0 }, { "off",   0 },
    { NULL,    0 }
  };
  int i;
  
  if(val) {
    for(i = 0; bools[i].str != NULL; i++) {
      if(!(strcasecmp(bools[i].str, val)))
	return bools[i].value;
    }
  }
  
  return 0;
}

/*
 * return argument number.
 */
int toxine_is_args(toxine_t *tox) {
  if(!tox)
    return -1;

  return tox->command.num_args;
}

/*
 * return argumemnt <num>, NULL if there is not argument.
 */
const char *toxine_get_arg(toxine_t *tox, int num) {
  if((tox == NULL) || (num < 1) || (tox->command.num_args < num))
    return NULL;
  
  return(tox->command.args[num - 1]);
}

/*
 * return 1 if *arg match with argument <pos>
 */
int toxine_is_arg_contain(toxine_t *tox, int pos, const char *arg) {
  
  if(tox && pos && ((arg != NULL) && (strlen(arg))) && (tox->command.num_args >= pos)) {
    if(!strncmp(tox->command.args[pos - 1], arg, strlen(tox->command.args[pos - 1])))
      return 1;
  }
  
  return 0;
}

/*
 * Set current command line from line.
 */
void toxine_set_command_line(toxine_t *tox, char *line) {

  if(tox && line && strlen(line)) {

    if(tox->command.line)
      tox->command.line = (char *) realloc(tox->command.line, strlen(line) + 1);
    else
      tox->command.line = (char *) xine_xmalloc(strlen(line) + 1);
    
    memset(tox->command.line, 0, sizeof(tox->command.line));
    sprintf(tox->command.line, "%s", line);

  }
}

/*
 * Look for first entry in playlist matching with 'mrl'
 */
static int toxine_lookup_in_playlist(toxine_t *tox, const char *mrl) {
  int i;

  if(tox->playlist.num) {
    for(i = 0; i < tox->playlist.num; i++) {
      if(!strcmp(mrl, tox->playlist.mmk[i]->mrl))
	return i;
    }
  }

  return -1;
}

/*
 *
 */
void toxine_set_current_mrl_from_cur(toxine_t *tox) {
  if(tox) {
    if((tox->playlist.cur >= 0) 
       && tox->playlist.mmk
       && tox->playlist.mmk[tox->playlist.cur]
       && (tox->playlist.mmk[tox->playlist.cur]->mrl != NULL))
      tox->current_mrl = tox->playlist.mmk[tox->playlist.cur]->mrl;
    else {
      tox->current_mrl = NULL;
      tox->playlist.cur = -1;
    }
  }
}

/*
 * Set current mrl from mrl
 */
void toxine_set_current_mrl(toxine_t *tox, const char *mrl) {

  if(tox && mrl) {

    if((tox->playlist.cur = toxine_lookup_in_playlist(tox, mrl)) == -1) {
      char *expanded_mrl;
      
      expanded_mrl = tilde_expand(mrl);
      
      playlist_add_entry(tox, expanded_mrl, NULL, 0, -1);
      tox->playlist.cur = (tox->playlist.num - 1);
      
      free(expanded_mrl);
    }
    
    toxine_set_current_mrl_from_cur(tox);
  }
}

/*
 * Execute a shell command.
 */
int toxine_system(int dont_run_as_root, char *command) {
  int pid, status;
  
  /* 
   * Don't permit run as root
   */
  if(dont_run_as_root) {
    if(getuid() == 0)
      return -1;
  }

  if(command == 0)
    return 1;

  pid = fork();

  if(pid == -1)
    return -1;
  
  if(pid == 0) {
    char *argv[4];
    argv[0] = "sh";
    argv[1] = "-c";
    argv[2] = command;
    argv[3] = 0;
    execve("/bin/sh", argv, environ);
    exit(127);
  }

  do {
    if(waitpid(pid, &status, 0) == -1) {
      if (errno != EINTR)
	return -1;
    } 
    else
      return WEXITSTATUS(status);

  } while(1);
  
  return -1;
}

/*
 *
 */
static int _mkdir_safe(char *path) {
  struct stat  pstat;
  
  if(path == NULL)
    return 0;
  
  if((lstat(path, &pstat)) < 0) {
    /* file or directory no exist, create it */
    if(mkdir(path, 0755) < 0) {
      fprintf(stderr, "mkdir(%s) failed: %s\n", path, strerror(errno));
      return 0;
    }
  }
  else {
    /* Check of found file is a directory file */
    if(!S_ISDIR(pstat.st_mode)) {
      fprintf(stderr, "%s is not a directory.\n", path);
      return 0;
    }
  }

  return 1;
}

/*
 *
 */
int toxine_mkdir_safe(char *path) {
  char *p, *pp;
  char  buf[_PATH_MAX + _NAME_MAX + 1];
  char  buf2[_PATH_MAX + _NAME_MAX + 1];
  
  if(path == NULL)
    return 0;
  
  memset(&buf, 0, sizeof(buf));
  memset(&buf2, 0, sizeof(buf2));
  
  sprintf(buf, "%s", path);
  pp = buf;
  while((p = xine_strsep(&pp, "/")) != NULL) {
    if(p && strlen(p)) {
      sprintf(buf2, "%s/%s", buf2, p);
      if(!_mkdir_safe(buf2))
	return 0;
    }
  }

  return 1;
}

/*
 *
 */
void toxine_set_last_char_result(toxine_t *tox, const char *result) {
  if(tox) {
    if(result) {
      if(tox->last_result)
	tox->last_result = (char *) realloc(tox->last_result, (strlen(result) + 1));
      else
	tox->last_result = (char *) xine_xmalloc(strlen(result) + 1);
      
      memset(tox->last_result, 0, sizeof(tox->last_result));
      sprintf(tox->last_result, "%s", result);
    }
    else {
      free(tox->last_result);
      tox->last_result = NULL;
    }
  }
}

void toxine_set_last_achar_result(toxine_t *tox, const char *const *result) {
  if(tox) {
    if(result) {
      int i = 0;
      int len = 0;
      
      while(result[i] != NULL)
	len += strlen(result[i++]);
      
      len += ((2 * (i - 1)) + 1);
      
      if(tox->last_result)
	tox->last_result = (char *) realloc(tox->last_result, (len + 1));
      else
	tox->last_result = (char *) xine_xmalloc(len + 1);
      
      memset(tox->last_result, 0, sizeof(tox->last_result));
      
      i = 0;
      while(result[i] != NULL) {
	if(i > 0)
	  strcat(tox->last_result, ", ");
	strcat(tox->last_result, result[i++]);
      }
    }
    else {
      free(tox->last_result);
      tox->last_result = NULL;
    }
  }
}

void toxine_set_last_int_result(toxine_t *tox, int result) {
  if(tox) {
    char buffer[132];

    memset(&buffer, 0, sizeof(buffer));
    sprintf(buffer, "%d", result);
    toxine_set_last_char_result(tox, buffer);
  }
}

char *toxine_get_last_char_result(toxine_t *tox) {
  if(tox)
    return tox->last_result;

  return NULL;
}

int toxine_get_last_int_result(toxine_t *tox) {
  int retval = 0;
  
  if(tox && (tox->last_result)) {
    char *p = tox->last_result;
    
    retval = strtol(p, &p, 10);
  }

  return 0;
}

void toxine_unset_last_result(toxine_t *tox) {
  if(tox) {
    if(tox->last_result) {
      free(tox->last_result);
      tox->last_result = NULL;
    }
  }
}

static void _config_update(toxine_t *tox,
			   _cfg_entry_t *entry, 
			   int type, int min, int max, int value, char *string) {

  switch(type) {

  case XINE_CONFIG_TYPE_UNKNOWN:
    fprintf(stderr, "Config key '%s' isn't registered yet.\n", entry->key);
    return;
    break;

  case XINE_CONFIG_TYPE_RANGE:
    entry->range_min = min;
    entry->range_max = max;
    break;

  case XINE_CONFIG_TYPE_STRING: 
    entry->str_value = string;
    break;
    
  case XINE_CONFIG_TYPE_ENUM:
  case XINE_CONFIG_TYPE_NUM:
  case XINE_CONFIG_TYPE_BOOL:
    entry->num_value = value;
    break;

  default:
    fprintf(stderr, "Unknown config type %d\n", type);
    return;
    break;
  }
  
  xine_config_update_entry(tox->xine, entry);
}

void toxine_config_update_range(toxine_t *tox, char *key, int min, int max) {
  _cfg_entry_t entry;

  memset(&entry, 0, sizeof(_cfg_entry_t));
  if(xine_config_lookup_entry(tox->xine, key, &entry)) 
    _config_update(tox, &entry, XINE_CONFIG_TYPE_RANGE, min, max, 0, NULL);
  else
    fprintf(stderr, "WOW, range key %s isn't registered\n", key);
}

void toxine_config_update_string(toxine_t *tox, char *key, char *string) {
  _cfg_entry_t entry;

  memset(&entry, 0, sizeof(_cfg_entry_t));
  
  if((xine_config_lookup_entry(tox->xine, key, &entry)) && string)
    _config_update(tox, &entry, XINE_CONFIG_TYPE_STRING, 0, 0, 0, string);
  else {
    if(string == NULL)
      fprintf(stderr, "string is NULL\n");
    else
      fprintf(stderr, "WOW, string key %s isn't registered\n", key);
  }
}

void toxine_config_update_enum(toxine_t *tox, char *key, int value) {
  _cfg_entry_t entry;

  memset(&entry, 0, sizeof(_cfg_entry_t));
  if(xine_config_lookup_entry(tox->xine, key, &entry)) 
    _config_update(tox, &entry, XINE_CONFIG_TYPE_ENUM, 0, 0, value, NULL);
  else
    fprintf(stderr, "WOW, enum key %s isn't registered\n", key);
}

void toxine_config_update_bool(toxine_t *tox, char *key, int value) {
  _cfg_entry_t entry;

  memset(&entry, 0, sizeof(_cfg_entry_t));
  if(xine_config_lookup_entry(tox->xine, key, &entry)) 
    _config_update(tox, &entry, XINE_CONFIG_TYPE_BOOL, 0, 0, ((value > 0) ? 1 : 0), NULL);
  else
    fprintf(stderr, "WOW, bool key %s isn't registered\n", key);
}

void toxine_config_update_num(toxine_t *tox, char *key, int value) {
  _cfg_entry_t entry;

  memset(&entry, 0, sizeof(_cfg_entry_t));
  if(xine_config_lookup_entry(tox->xine, key, &entry)) 
    _config_update(tox, &entry, XINE_CONFIG_TYPE_NUM, 0, 0, value, NULL);
  else
    fprintf(stderr, "WOW, num key %s isn't registered\n", key);
}

void toxine_config_load(toxine_t *tox) {
  if(tox)
    xine_config_load(tox->xine, tox->configfile);
}

void toxine_config_save(toxine_t *tox) {
  if(tox) {
    return_if_no_new(tox);
    xine_config_save(tox->xine, tox->configfile);
  }
}

void toxine_show_stream_info(int info_type, uint32_t iinfo) {
  switch(info_type) {
  case XINE_STREAM_INFO_BITRATE:
  case XINE_STREAM_INFO_SEEKABLE:
  case XINE_STREAM_INFO_AUDIO_HANDLED:
  case XINE_STREAM_INFO_HAS_CHAPTERS:
  case XINE_STREAM_INFO_HAS_VIDEO:
  case XINE_STREAM_INFO_HAS_AUDIO:
  case XINE_STREAM_INFO_IGNORE_VIDEO:
  case XINE_STREAM_INFO_IGNORE_AUDIO:
  case XINE_STREAM_INFO_IGNORE_SPU:
  case XINE_STREAM_INFO_VIDEO_HAS_STILL:
    pinfo("%s\n", toxine_get_yesno_string(iinfo));
    break;
    
  case XINE_STREAM_INFO_AUDIO_FOURCC:
  case XINE_STREAM_INFO_VIDEO_FOURCC:
    pinfo("%s\n", toxine_get_fourcc_string(iinfo));
    break;
    
  case XINE_STREAM_INFO_VIDEO_WIDTH:
  case XINE_STREAM_INFO_VIDEO_HEIGHT:
  case XINE_STREAM_INFO_VIDEO_RATIO:
  case XINE_STREAM_INFO_VIDEO_CHANNELS:
  case XINE_STREAM_INFO_VIDEO_STREAMS:
  case XINE_STREAM_INFO_VIDEO_BITRATE:
  case XINE_STREAM_INFO_VIDEO_HANDLED:
  case XINE_STREAM_INFO_FRAME_DURATION:
  case XINE_STREAM_INFO_AUDIO_CHANNELS:
  case XINE_STREAM_INFO_AUDIO_BITS:
  case XINE_STREAM_INFO_AUDIO_SAMPLERATE:
  case XINE_STREAM_INFO_AUDIO_BITRATE:
  case XINE_STREAM_INFO_MAX_AUDIO_CHANNEL:
  case XINE_STREAM_INFO_MAX_SPU_CHANNEL:
    pinfo("%d\n", iinfo);
    break;

  default:
    perr("Unknown stream info type %d\n", iinfo);
    break;
  }
}
